@@ -26,7 +26,6 @@ from pointblank.datascan import DataScan, col_summary_tbl
 from pointblank.draft import DraftValidation
 from pointblank.schema import Schema
 from pointblank.segments import seg_group
-from pointblank.tf import TF
 from pointblank.thresholds import Actions, FinalActions, Thresholds
 from pointblank.validate import (
     Validate,
@@ -49,7 +48,6 @@ from pointblank.yaml import (
 
 __all__ = [
     "assistant",
-    "TF",
     "Validate",
     "Thresholds",
     "Actions",
