@@ -1,7 +1,8 @@
 from __future__ import annotations
 
+from importlib import import_module
 from pathlib import Path
-from typing import Any, Union
+from typing import Any, Iterable, Mapping, Optional, Union
 
 import yaml
 from narwhals.typing import FrameT
@@ -17,7 +18,9 @@ class YAMLValidationError(Exception):
     pass
 
 
-def _safe_eval_python_code(code: str) -> Any:
+def _safe_eval_python_code(
+    code: str, namespaces: Optional[Union[Iterable[str], Mapping[str, str]]] = None
+) -> Any:
     """Safely evaluate Python code with restricted namespace.
 
     This function provides a controlled environment for executing Python code embedded in YAML
@@ -68,6 +71,7 @@ def _safe_eval_python_code(code: str) -> Any:
             "abs": abs,
             "round": round,
             "print": print,
+            "__import__": __import__,
         },
     }
 
@@ -88,12 +92,25 @@ def _safe_eval_python_code(code: str) -> Any:
 
         safe_namespace["pd"] = pd
 
-    # Check for dangerous patterns
+    if namespaces:
+        for alias, module_name in (
+            namespaces.items() if isinstance(namespaces, dict) else ((m, m) for m in namespaces)
+        ):
+            try:
+                safe_namespace[alias] = import_module(module_name)
+            except ImportError as e:
+                raise ImportError(
+                    f"Could not import requested namespace '{module_name}': {e}"
+                ) from e
+
+    # Check for dangerous patterns and be more specific about __import__ to allow legitimate use
     dangerous_patterns = [
-        r"import\s+os",
-        r"import\s+sys",
-        r"import\s+subprocess",
-        r"__import__",
+        r"import\s+os\b",
+        r"import\s+sys\b",
+        r"import\s+subprocess\b",
+        r"__import__\s*\(\s*['\"]os['\"]",
+        r"__import__\s*\(\s*['\"]sys['\"]",
+        r"__import__\s*\(\s*['\"]subprocess['\"]",
         r"exec\s*\(",
         r"eval\s*\(",
         r"open\s*\(",
@@ -142,7 +159,9 @@ def _safe_eval_python_code(code: str) -> Any:
         raise YAMLValidationError(f"Error executing Python code '{code}': {e}")
 
 
-def _process_python_expressions(value: Any) -> Any:
+def _process_python_expressions(
+    value: Any, namespaces: Optional[Union[Iterable[str], Mapping[str, str]]] = None
+) -> Any:
     """Process Python code snippets embedded in YAML values.
 
     This function supports the python: block syntax for embedding Python code:
@@ -152,7 +171,7 @@ def _process_python_expressions(value: Any) -> Any:
       pl.scan_csv("data.csv").head(10)
 
     Note: col_vals_expr() also supports a shortcut syntax where the expr parameter
-    can be written directly without the python: wrapper:
+    can be written directly without the python: wrapper: +
 
     col_vals_expr:
       expr: |
@@ -180,14 +199,14 @@ def _process_python_expressions(value: Any) -> Any:
         # Handle python: block syntax
         if "python" in value and len(value) == 1:
             code = value["python"]
-            return _safe_eval_python_code(code)
+            return _safe_eval_python_code(code, namespaces=namespaces)
 
         # Recursively process dictionary values
-        return {k: _process_python_expressions(v) for k, v in value.items()}
+        return {k: _process_python_expressions(v, namespaces=namespaces) for k, v in value.items()}
 
     elif isinstance(value, list):
         # Recursively process list items
-        return [_process_python_expressions(item) for item in value]
+        return [_process_python_expressions(item, namespaces=namespaces) for item in value]
 
     else:
         # Return primitive types unchanged
@@ -302,7 +321,7 @@ class YAMLValidator:
             raise YAMLValidationError("YAML must contain 'steps' field")
 
         if not isinstance(config["steps"], list):
-            raise YAMLValidationError("'steps' must be a list")
+            raise YAMLValidationError("'steps' must be a list")  # pragma: no cover
 
         if len(config["steps"]) == 0:
             raise YAMLValidationError("'steps' cannot be empty")
@@ -393,9 +412,9 @@ class YAMLValidator:
             if processed_data is processed_tbl_spec and isinstance(processed_tbl_spec, str):
                 return load_dataset(processed_tbl_spec, tbl_type=df_library)
             else:
-                return processed_data
+                return processed_data  # pragma: no cover
 
-        except Exception as e:
+        except Exception as e:  # pragma: no cover
             raise YAMLValidationError(f"Failed to load data source '{tbl_spec}': {e}")
 
     def _load_csv_file(self, file_path: str, df_library: str) -> Any:
@@ -439,16 +458,16 @@ class YAMLValidator:
 
             elif df_library == "duckdb":
                 # For DuckDB, we'll use the existing _process_data since it handles DuckDB
-                from pointblank.validate import _process_data
+                from pointblank.validate import _process_data  # pragma: no cover
 
-                return _process_data(file_path)
+                return _process_data(file_path)  # pragma: no cover
 
             else:
                 raise YAMLValidationError(
                     f"Unsupported df_library: {df_library}. Use 'polars', 'pandas', or 'duckdb'"
                 )
 
-        except Exception as e:
+        except Exception as e:  # pragma: no cover
             raise YAMLValidationError(
                 f"Failed to load CSV file '{file_path}' with {df_library}: {e}"
             )
@@ -547,7 +566,11 @@ class YAMLValidator:
                 f"Schema specification must be a dictionary, got: {type(schema_spec)}"
             )
 
-    def _parse_validation_step(self, step_config: Union[str, dict]) -> tuple[str, dict]:
+    def _parse_validation_step(
+        self,
+        step_config: Union[str, dict],
+        namespaces: Optional[Union[Iterable[str], Mapping[str, str]]] = None,
+    ) -> tuple[str, dict]:
         """Parse a single validation step from YAML configuration.
 
         Parameters
@@ -598,14 +621,16 @@ class YAMLValidator:
             # Special case: `col_vals_expr()`'s `expr=` parameter can use shortcut syntax
             if method_name == "col_vals_expr" and key == "expr" and isinstance(value, str):
                 # Treat string directly as Python code (shortcut syntax)
-                processed_parameters[key] = _safe_eval_python_code(value)
+                processed_parameters[key] = _safe_eval_python_code(value, namespaces=namespaces)
             # Special case: `pre=` parameter can use shortcut syntax (like `expr=`)
             elif key == "pre" and isinstance(value, str):
                 # Treat string directly as Python code (shortcut syntax)
-                processed_parameters[key] = _safe_eval_python_code(value)
+                processed_parameters[key] = _safe_eval_python_code(value, namespaces=namespaces)
             else:
                 # Normal processing (requires python: block syntax)
-                processed_parameters[key] = _process_python_expressions(value)
+                processed_parameters[key] = _process_python_expressions(
+                    value, namespaces=namespaces
+                )
         parameters = processed_parameters
 
         # Convert `columns=` specification
@@ -634,7 +659,7 @@ class YAMLValidator:
                     if isinstance(expr, str):
                         lambda_expressions.append(_safe_eval_python_code(expr))
                     else:
-                        lambda_expressions.append(expr)
+                        lambda_expressions.append(expr)  # pragma: no cover
                 # Pass expressions as positional arguments (stored as special key)
                 parameters["_conjointly_expressions"] = lambda_expressions
             else:
@@ -658,7 +683,9 @@ class YAMLValidator:
 
         return self.validation_method_map[method_name], parameters
 
-    def build_validation(self, config: dict) -> Validate:
+    def build_validation(
+        self, config: dict, namespaces: Optional[Union[Iterable[str], Mapping[str, str]]] = None
+    ) -> Validate:
         """Convert YAML config to Validate object.
 
         Parameters
@@ -693,7 +720,9 @@ class YAMLValidator:
         # Set actions if provided
         if "actions" in config:
             # Process actions: handle `python:` block syntax for callables
-            processed_actions = _process_python_expressions(config["actions"])
+            processed_actions = _process_python_expressions(
+                config["actions"], namespaces=namespaces
+            )
             # Convert to Actions object
             validate_kwargs["actions"] = Actions(**processed_actions)
 
@@ -713,7 +742,9 @@ class YAMLValidator:
 
         # Add validation steps
         for step_config in config["steps"]:
-            method_name, parameters = self._parse_validation_step(step_config)
+            method_name, parameters = self._parse_validation_step(
+                step_config, namespaces=namespaces
+            )
 
             # Get the method from the validation object
             method = getattr(validation, method_name)
@@ -728,7 +759,9 @@ class YAMLValidator:
 
         return validation
 
-    def execute_workflow(self, config: dict) -> Validate:
+    def execute_workflow(
+        self, config: dict, namespaces: Optional[Union[Iterable[str], Mapping[str, str]]] = None
+    ) -> Validate:
         """Execute a complete YAML validation workflow.
 
         Parameters
@@ -742,7 +775,7 @@ class YAMLValidator:
             Interrogated Validate object with results.
         """
         # Build the validation plan
-        validation = self.build_validation(config)
+        validation = self.build_validation(config, namespaces=namespaces)
 
         # Execute interrogation to get results
         validation = validation.interrogate()
@@ -750,7 +783,11 @@ class YAMLValidator:
         return validation
 
 
-def yaml_interrogate(yaml: Union[str, Path], set_tbl: Union[FrameT, Any, None] = None) -> Validate:
+def yaml_interrogate(
+    yaml: Union[str, Path],
+    set_tbl: Union[FrameT, Any, None] = None,
+    namespaces: Optional[Union[Iterable[str], Mapping[str, str]]] = None,
+) -> Validate:
     """Execute a YAML-based validation workflow.
 
     This is the main entry point for YAML-based validation workflows. It takes YAML configuration
@@ -772,6 +809,10 @@ def yaml_interrogate(yaml: Union[str, Path], set_tbl: Union[FrameT, Any, None] =
         `tbl` field before executing the validation workflow. This can be any supported table type
         including DataFrame objects, Ibis table objects, CSV file paths, Parquet file paths, GitHub
         URLs, or database connection strings.
+    namespaces
+        Optional module namespaces to make available for Python code execution in YAML
+        configurations. Can be a dictionary mapping aliases to module names or a list of module
+        names. See the "Using Namespaces" section below for detailed examples.
 
     Returns
     -------
@@ -786,6 +827,71 @@ def yaml_interrogate(yaml: Union[str, Path], set_tbl: Union[FrameT, Any, None] =
         If the YAML is invalid, malformed, or execution fails. This includes syntax errors, missing
         required fields, unknown validation methods, or data loading failures.
 
+    Using Namespaces
+    ----------------
+    The `namespaces=` parameter enables custom Python modules and functions in YAML configurations.
+    This is particularly useful for custom action functions and advanced Python expressions.
+
+    **Namespace formats:**
+
+    - Dictionary format: `{"alias": "module.name"}` maps aliases to module names
+    - List format: `["module.name", "another.module"]` imports modules directly
+
+    **Option 1: Inline expressions (no namespaces needed)**
+
+    ```{python}
+    import pointblank as pb
+
+    # Simple inline custom action
+    yaml_config = '''
+    tbl: small_table
+    thresholds:
+      warning: 0.01
+    actions:
+      warning:
+        python: "lambda: print('Custom warning triggered')"
+    steps:
+    - col_vals_gt:
+        columns: [a]
+        value: 1000
+    '''
+
+    result = pb.yaml_interrogate(yaml_config)
+    result
+    ```
+
+    **Option 2: External functions with namespaces**
+
+    ```{python}
+    # Define a custom action function
+    def my_custom_action():
+        print("Data validation failed: please check your data.")
+
+    # Add to current module for demo
+    import sys
+    sys.modules[__name__].my_custom_action = my_custom_action
+
+    # YAML that references the external function
+    yaml_config = '''
+    tbl: small_table
+    thresholds:
+      warning: 0.01
+    actions:
+      warning:
+        python: actions.my_custom_action
+    steps:
+    - col_vals_gt:
+        columns: [a]
+        value: 1000  # This will fail
+    '''
+
+    # Use namespaces to make the function available
+    result = pb.yaml_interrogate(yaml_config, namespaces={'actions': '__main__'})
+    result
+    ```
+
+    This approach enables modular, reusable validation workflows with custom business logic.
+
     Examples
     --------
     ```{python}
@@ -928,14 +1034,14 @@ def yaml_interrogate(yaml: Union[str, Path], set_tbl: Union[FrameT, Any, None] =
     # If `set_tbl=` is provided, we need to build the validation workflow and then use `set_tbl()`
     if set_tbl is not None:
         # First build the validation object without interrogation
-        validation = validator.build_validation(config)
+        validation = validator.build_validation(config, namespaces=namespaces)
         # Then replace the table using set_tbl method
         validation = validation.set_tbl(tbl=set_tbl)
         # Finally interrogate with the new table
         return validation.interrogate()
     else:
         # Standard execution without table override (includes interrogation)
-        return validator.execute_workflow(config)
+        return validator.execute_workflow(config, namespaces=namespaces)
 
 
 def load_yaml_config(file_path: Union[str, Path]) -> dict:
@@ -1223,7 +1329,7 @@ def yaml_to_python(yaml: Union[str, Path]) -> str:
     """
     # First, parse the raw YAML to detect Polars/Pandas expressions in the source code
     if isinstance(yaml, Path):
-        yaml_content = yaml.read_text()
+        yaml_content = yaml.read_text()  # pragma: no cover
     elif isinstance(yaml, str):
         # Check if it's a file path (single line, reasonable length, no newlines)
         if len(yaml) < 260 and "\n" not in yaml and Path(yaml).exists():
@@ -1231,7 +1337,7 @@ def yaml_to_python(yaml: Union[str, Path]) -> str:
         else:
             yaml_content = yaml
     else:
-        yaml_content = str(yaml)
+        yaml_content = str(yaml)  # pragma: no cover
 
     # Track whether we need to import Polars and Pandas by analyzing the raw YAML content
     needs_polars_import = False
@@ -1326,7 +1432,7 @@ def yaml_to_python(yaml: Union[str, Path]) -> str:
             validate_args.append(f'data=pb.load_dataset("{tbl_spec}", tbl_type="{df_library}")')
     else:
         # Fallback to placeholder if we couldn't extract the original expression
-        validate_args.append("data=<python_expression_result>")
+        validate_args.append("data=<python_expression_result>")  # pragma: no cover
 
     # Add table name if present
     if "tbl_name" in config:
@@ -1359,7 +1465,7 @@ def yaml_to_python(yaml: Union[str, Path]) -> str:
                 action_params.append(f'{key}="{value}"')
             else:
                 # For callables or complex expressions, use placeholder
-                action_params.append(f"{key}={value}")
+                action_params.append(f"{key}={value}")  # pragma: no cover
         actions_str = "pb.Actions(" + ", ".join(action_params) + ")"
         validate_args.append(f"actions={actions_str}")
 
@@ -1414,7 +1520,7 @@ def yaml_to_python(yaml: Union[str, Path]) -> str:
             elif isinstance(step_params["expr"], str):
                 original_expressions["expr"] = step_params["expr"]
 
-        method_name, parameters = validator._parse_validation_step(step_config)
+        method_name, parameters = validator._parse_validation_step(step_config, namespaces=None)
 
         # Apply the original expressions to override the converted lambda functions
         if method_name == "conjointly" and "expressions" in original_expressions:
@@ -1446,13 +1552,13 @@ def yaml_to_python(yaml: Union[str, Path]) -> str:
                     expressions_str = "[" + ", ".join([f'"{expr}"' for expr in value]) + "]"
                     param_parts.append(f"expressions={expressions_str}")
                 else:
-                    param_parts.append(f"expressions={value}")
+                    param_parts.append(f"expressions={value}")  # pragma: no cover
             elif key == "expr" and method_name == "specially":
                 # Handle specially expr parameter: should be unquoted lambda expression
                 if isinstance(value, str):
                     param_parts.append(f"expr={value}")
                 else:
-                    param_parts.append(f"expr={value}")
+                    param_parts.append(f"expr={value}")  # pragma: no cover
             elif key in ["columns", "columns_subset"]:
                 if isinstance(value, list):
                     if len(value) == 1:
@@ -1463,7 +1569,7 @@ def yaml_to_python(yaml: Union[str, Path]) -> str:
                         columns_str = "[" + ", ".join([f'"{col}"' for col in value]) + "]"
                         param_parts.append(f"{key}={columns_str}")
                 else:
-                    param_parts.append(f'{key}="{value}"')
+                    param_parts.append(f'{key}="{value}"')  # pragma: no cover
             elif key == "brief":
                 # Handle `brief=` parameter: can be a boolean or a string
                 if isinstance(value, bool):
@@ -1486,50 +1592,34 @@ def yaml_to_python(yaml: Union[str, Path]) -> str:
                         elif isinstance(value.warning, list) and len(value.warning) == 1:
                             action_params.append(f'warning="{value.warning[0]}"')
                         else:
-                            action_params.append(f"warning={value.warning}")
+                            action_params.append(f"warning={value.warning}")  # pragma: no cover
 
                     if value.error is not None:
                         error_expr_path = f"{step_action_base}.error"
                         if error_expr_path in step_expressions:
-                            action_params.append(f"error={step_expressions[error_expr_path]}")
+                            action_params.append(
+                                f"error={step_expressions[error_expr_path]}"
+                            )  # pragma: no cover
                         elif isinstance(value.error, list) and len(value.error) == 1:
                             action_params.append(f'error="{value.error[0]}"')
                         else:
-                            action_params.append(f"error={value.error}")
+                            action_params.append(f"error={value.error}")  # pragma: no cover
 
                     if value.critical is not None:
                         critical_expr_path = f"{step_action_base}.critical"
                         if critical_expr_path in step_expressions:
-                            action_params.append(f"critical={step_expressions[critical_expr_path]}")
+                            action_params.append(
+                                f"critical={step_expressions[critical_expr_path]}"
+                            )  # pragma: no cover
                         elif isinstance(value.critical, list) and len(value.critical) == 1:
                             action_params.append(f'critical="{value.critical[0]}"')
                         else:
-                            action_params.append(f"critical={value.critical}")
+                            action_params.append(f"critical={value.critical}")  # pragma: no cover
 
                     if hasattr(value, "highest_only") and value.highest_only is not True:
                         action_params.append(f"highest_only={value.highest_only}")
                     actions_str = "pb.Actions(" + ", ".join(action_params) + ")"
                     param_parts.append(f"actions={actions_str}")
-                elif isinstance(value, dict):
-                    action_params = []
-                    step_action_base = f"steps[{step_index}].{list(step_config.keys())[0]}.actions"
-                    for action_key, action_value in value.items():
-                        if action_key == "highest_only":
-                            action_params.append(f"{action_key}={action_value}")
-                        else:
-                            # Check if we have an original expression for this action
-                            action_expr_path = f"{step_action_base}.{action_key}"
-                            if action_expr_path in step_expressions:
-                                action_params.append(
-                                    f"{action_key}={step_expressions[action_expr_path]}"
-                                )
-                            elif isinstance(action_value, str):
-                                action_params.append(f'{action_key}="{action_value}"')
-                            else:
-                                # For callables or complex expressions
-                                action_params.append(f"{action_key}={action_value}")
-                    actions_str = "pb.Actions(" + ", ".join(action_params) + ")"
-                    param_parts.append(f"actions={actions_str}")
                 else:
                     param_parts.append(f"actions={value}")
             elif key == "thresholds":
