@@ -13,8 +13,11 @@ from pointblank._utils import (
     _check_column_type,
     _check_invalid_fields,
     _column_test_prep,
+    _column_subset_test_prep,
     _count_true_values_in_column,
     _count_null_values_in_column,
+    _derive_single_bound,
+    _derive_bounds,
     _format_to_float_value,
     _format_to_integer_value,
     _get_assertion_from_fname,
@@ -24,10 +27,18 @@ from pointblank._utils import (
     _get_examples_text,
     _get_fn_name,
     _get_tbl_type,
+    _is_lazy_frame,
+    _is_lib_present,
+    _is_narwhals_table,
     _is_numeric_dtype,
     _is_date_or_datetime_dtype,
     _is_duration_dtype,
+    _is_value_a_df,
+    _check_any_df_lib,
+    _pivot_to_dict,
+    _process_ibis_through_narwhals,
     _select_df_lib,
+    transpose_dicts,
 )
 
 from pointblank.validate import load_dataset
@@ -509,3 +520,240 @@ def test_get_examples_text():
 
 def test_get_api_and_examples_text():
     assert isinstance(_get_api_and_examples_text(), str)
+
+
+def test_transpose_dicts():
+    # Test with empty list
+    assert transpose_dicts([]) == {}
+
+    # Test with single dict
+    result = transpose_dicts([{"a": 1, "b": 2}])
+    expected = {"a": [1], "b": [2]}
+    assert result == expected
+
+    # Test with multiple dicts with same keys
+    result = transpose_dicts([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
+    expected = {"a": [1, 3], "b": [2, 4]}
+    assert result == expected
+
+    # Test with multiple dicts with different keys (missing values become None)
+    result = transpose_dicts([{"a": 1, "b": 2}, {"a": 3, "c": 4}])
+    expected = {"a": [1, 3], "b": [2, None], "c": [None, 4]}
+    assert result == expected
+
+
+def test_derive_single_bound():
+    # Test with valid inputs
+    assert _derive_single_bound(100, 0.1) == 10  # tol < 1, so tol * ref
+    assert _derive_single_bound(100, 1.5) == 1  # tol >= 1, so int(tol)
+    assert _derive_single_bound(100, 2) == 2  # integer tolerance
+    assert _derive_single_bound(50, 0.5) == 25  # half of reference
+
+    # Test error cases
+    with pytest.raises(TypeError):
+        _derive_single_bound(100, "invalid")
+
+    with pytest.raises(TypeError):
+        _derive_single_bound(100, [1, 2])
+
+    with pytest.raises(ValueError):
+        _derive_single_bound(100, -1)
+
+    with pytest.raises(ValueError):
+        _derive_single_bound(100, -0.5)
+
+
+def test_derive_bounds():
+    # Test with single tolerance value
+    result = _derive_bounds(100, 0.1)
+    expected = (10, 10)
+    assert result == expected
+
+    # Test with tuple tolerance
+    result = _derive_bounds(100, (0.1, 0.2))
+    expected = (10, 20)
+    assert result == expected
+
+    # Test with tuple of integers
+    result = _derive_bounds(100, (2, 3))
+    expected = (2, 3)
+    assert result == expected
+
+
+def test_is_narwhals_table():
+    # Test with actual Polars/Pandas DataFrames (not narwhals wrapped)
+    pd_df = pd.DataFrame({"x": [1, 2, 3]})
+    pl_df = pl.DataFrame({"x": [1, 2, 3]})
+    assert not _is_narwhals_table(pd_df)
+    assert not _is_narwhals_table(pl_df)
+
+    # Test with narwhals wrapped DataFrame
+    nw_df = nw.from_native(pd_df)
+    assert _is_narwhals_table(nw_df)
+
+    # Test with non-DataFrame objects
+    assert not _is_narwhals_table("string")
+    assert not _is_narwhals_table(123)
+    assert not _is_narwhals_table([1, 2, 3])
+
+
+def test_is_lazy_frame():
+    # Test with regular DataFrames
+    pd_df = pd.DataFrame({"x": [1, 2, 3]})
+    pl_df = pl.DataFrame({"x": [1, 2, 3]})
+    assert not _is_lazy_frame(pd_df)
+    assert not _is_lazy_frame(pl_df)
+
+    # Test with lazy frame
+    pl_lazy = pl.DataFrame({"x": [1, 2, 3]}).lazy()
+    assert _is_lazy_frame(pl_lazy)
+
+    # Test with narwhals lazy frame
+    nw_lazy = nw.from_native(pl_lazy)
+    assert _is_lazy_frame(nw_lazy)
+
+    # Test with non-DataFrame objects
+    assert not _is_lazy_frame("string")
+    assert not _is_lazy_frame(123)
+
+
+def test_is_lib_present():
+    # Test with libraries that should be present
+    assert _is_lib_present("sys")
+    assert _is_lib_present("os")
+    assert _is_lib_present("pandas")
+    assert _is_lib_present("polars")
+
+    # Test with library that should not be present
+    assert not _is_lib_present("nonexistent_library")
+
+
+def test_check_any_df_lib():
+    # Should not raise an error since both pandas and polars are available
+    _check_any_df_lib("test_method")
+
+    # Test error case when neither pandas nor polars is available
+    with patch.dict(sys.modules, {"pandas": None, "polars": None}):
+        with pytest.raises(ImportError):
+            _check_any_df_lib("test_method")
+
+
+def test_is_value_a_df():
+    # Test with valid DataFrames
+    pd_df = pd.DataFrame({"x": [1, 2, 3]})
+    pl_df = pl.DataFrame({"x": [1, 2, 3]})
+    assert _is_value_a_df(pd_df)
+    assert _is_value_a_df(pl_df)
+
+    # Test with non-DataFrame objects
+    assert not _is_value_a_df("string")
+    assert not _is_value_a_df(123)
+    assert not _is_value_a_df([1, 2, 3])
+    assert not _is_value_a_df({"x": [1, 2, 3]})
+
+
+@pytest.mark.parametrize(
+    "tbl_fixture",
+    ["tbl_multiple_types_pd", "tbl_multiple_types_pl"],
+)
+def test_column_subset_test_prep(request, tbl_fixture):
+    tbl = request.getfixturevalue(tbl_fixture)
+
+    # Test with valid column subset
+    dfn = _column_subset_test_prep(df=tbl, columns_subset=["int", "str"])
+    assert isinstance(dfn, nw.DataFrame)
+
+    # Test with None columns_subset
+    dfn = _column_subset_test_prep(df=tbl, columns_subset=None)
+    assert isinstance(dfn, nw.DataFrame)
+
+    # Test with empty columns_subset
+    dfn = _column_subset_test_prep(df=tbl, columns_subset=[])
+    assert isinstance(dfn, nw.DataFrame)
+
+    # Test with check_exists=False (should not raise even with invalid columns)
+    dfn = _column_subset_test_prep(df=tbl, columns_subset=["invalid"], check_exists=False)
+    assert isinstance(dfn, nw.DataFrame)
+
+    # Test error case with invalid column when check_exists=True
+    with pytest.raises(ValueError):
+        _column_subset_test_prep(df=tbl, columns_subset=["invalid"], check_exists=True)
+
+
+def test_pivot_to_dict():
+    # Test basic functionality
+    input_dict = {
+        "col1": {"key1": "value1", "key2": "value2"},
+        "col2": {"key1": "value3", "key3": "value4"},
+    }
+
+    result = _pivot_to_dict(input_dict)
+    expected = {"key1": ["value1", "value3"], "key2": ["value2", None], "key3": [None, "value4"]}
+    assert result == expected
+
+    # Test with empty dict
+    assert _pivot_to_dict({}) == {}
+
+    # Test with single column
+    input_dict = {"col1": {"key1": "value1", "key2": "value2"}}
+    result = _pivot_to_dict(input_dict)
+    expected = {"key1": ["value1"], "key2": ["value2"]}
+    assert result == expected
+
+
+def test_process_ibis_through_narwhals():
+    # Test with non-Ibis table type
+    pd_df = pd.DataFrame({"x": [1, 2, 3]})
+    result_data, result_type = _process_ibis_through_narwhals(pd_df, "pandas")
+    assert result_data is pd_df
+    assert result_type == "pandas"
+
+    # Test with mock Ibis table type
+    # Since we can't easily create real Ibis tables in tests, we'll mock the behavior
+    with patch("pointblank._utils.nw.from_native") as mock_nw:
+        mock_nw.return_value = "mocked_narwhals_df"
+
+        # Test successful Narwhals wrapping
+        result_data, result_type = _process_ibis_through_narwhals("mock_ibis_table", "duckdb")
+        assert result_data == "mocked_narwhals_df"
+        assert result_type == "narwhals"
+
+        # Test fallback when Narwhals fails
+        mock_nw.side_effect = Exception("Narwhals failed")
+        result_data, result_type = _process_ibis_through_narwhals("mock_ibis_table", "duckdb")
+        assert result_data == "mock_ibis_table"
+        assert result_type == "duckdb"
+
+
+def test_get_tbl_type_additional():
+    # Test with invalid data that raises TypeError
+    with pytest.raises(TypeError):
+        _get_tbl_type("invalid_data")
+
+
+def test_get_tbl_type_pyspark():
+    # Test pyspark detection by mocking narwhals
+    class MockPySparkNamespace:
+        def __str__(self):
+            return "pyspark.sql.module"
+
+    class MockPySparkDF:
+        def __native_namespace__(self):
+            return MockPySparkNamespace()
+
+    mock_pyspark_df = MockPySparkDF()
+
+    with patch("pointblank._utils.nw.from_native", return_value=mock_pyspark_df):
+        # Create a non-Ibis object to trigger the regular DataFrame path
+        mock_data = type("MockNonIbisData", (), {})()
+        result = _get_tbl_type(mock_data)
+        assert result == "pyspark"
+
+
+def test_get_column_dtype_raw_parameter():
+    tbl = pd.DataFrame({"int_col": [1, 2, 3]})
+    dfn = _convert_to_narwhals(tbl)
+
+    # Test raw=True - this should return the raw dtype from the schema
+    raw_dtype = _get_column_dtype(dfn=dfn, column="int_col", raw=True)
+    assert raw_dtype is not None  # Just check that it returns something
